@@ -7,12 +7,13 @@ sticking incoming connections onto a Queue::
 
     server = HTTPServer(...)
     server.start()
-    ->  while True:
-            tick()
-            # This blocks until a request comes in:
-            child = socket.accept()
-            conn = HTTPConnection(child, ...)
-            server.requests.put(conn)
+    ->  serve()
+        while ready:
+            _connections.run()
+                while not stop_requested:
+                    child = socket.accept()  # blocks until a request comes in
+                    conn = HTTPConnection(child, ...)
+                    server.process_conn(conn)  # adds conn to threadpool
 
 Worker threads are kept in a pool and poll the Queue, popping off and then
 handling each connection in turn. Each connection can consist of an arbitrary
@@ -56,6 +57,10 @@ will run the server forever) or use invoking :func:`prepare()
 
 And now for a trivial doctest to exercise the test suite
 
+.. testsetup::
+
+   from cheroot.server import HTTPServer
+
 >>> 'HTTPServer' in globals()
 True
 """
@@ -75,7 +80,6 @@ import logging
 import platform
 import contextlib
 import threading
-import errno
 
 try:
     from functools import lru_cache
@@ -162,6 +166,9 @@ QUOTED_SLASH = b'%2F'
 QUOTED_SLASH_REGEX = re.compile(b''.join((b'(?i)', QUOTED_SLASH)))
 
 
+_STOPPING_FOR_INTERRUPT = object()  # sentinel used during shutdown
+
+
 comma_separated_headers = [
     b'Accept', b'Accept-Charset', b'Accept-Encoding',
     b'Accept-Language', b'Accept-Ranges', b'Allow', b'Cache-Control',
@@ -270,7 +277,8 @@ class SizeCheckWrapper:
     def read(self, size=None):
         """Read a chunk from ``rfile`` buffer and return it.
 
-        :param int size: amount of data to read
+        :param size: amount of data to read
+        :type size: int
 
         :returns: chunk from ``rfile``, limited by size if specified
         :rtype: bytes
@@ -283,7 +291,8 @@ class SizeCheckWrapper:
     def readline(self, size=None):
         """Read a single line from ``rfile`` buffer and return it.
 
-        :param int size: minimum amount of data to read
+        :param size: minimum amount of data to read
+        :type size: int
 
         :returns: one line from ``rfile``
         :rtype: bytes
@@ -309,7 +318,8 @@ class SizeCheckWrapper:
     def readlines(self, sizehint=0):
         """Read all lines from ``rfile`` buffer and return them.
 
-        :param int sizehint: hint of minimum amount of data to read
+        :param sizehint: hint of minimum amount of data to read
+        :type sizehint: int
 
         :returns: lines of bytes read from ``rfile``
         :rtype: list[bytes]
@@ -359,7 +369,8 @@ class KnownLengthRFile:
     def read(self, size=None):
         """Read a chunk from ``rfile`` buffer and return it.
 
-        :param int size: amount of data to read
+        :param size: amount of data to read
+        :type size: int
 
         :rtype: bytes
         :returns: chunk from ``rfile``, limited by size if specified
@@ -378,7 +389,8 @@ class KnownLengthRFile:
     def readline(self, size=None):
         """Read a single line from ``rfile`` buffer and return it.
 
-        :param int size: minimum amount of data to read
+        :param size: minimum amount of data to read
+        :type size: int
 
         :returns: one line from ``rfile``
         :rtype: bytes
@@ -397,7 +409,8 @@ class KnownLengthRFile:
     def readlines(self, sizehint=0):
         """Read all lines from ``rfile`` buffer and return them.
 
-        :param int sizehint: hint of minimum amount of data to read
+        :param sizehint: hint of minimum amount of data to read
+        :type sizehint: int
 
         :returns: lines of bytes read from ``rfile``
         :rtype: list[bytes]
@@ -498,7 +511,8 @@ class ChunkedRFile:
     def read(self, size=None):
         """Read a chunk from ``rfile`` buffer and return it.
 
-        :param int size: amount of data to read
+        :param size: amount of data to read
+        :type size: int
 
         :returns: chunk from ``rfile``, limited by size if specified
         :rtype: bytes
@@ -529,7 +543,8 @@ class ChunkedRFile:
     def readline(self, size=None):
         """Read a single line from ``rfile`` buffer and return it.
 
-        :param int size: minimum amount of data to read
+        :param size: minimum amount of data to read
+        :type size: int
 
         :returns: one line from ``rfile``
         :rtype: bytes
@@ -570,7 +585,8 @@ class ChunkedRFile:
     def readlines(self, sizehint=0):
         """Read all lines from ``rfile`` buffer and return them.
 
-        :param int sizehint: hint of minimum amount of data to read
+        :param sizehint: hint of minimum amount of data to read
+        :type sizehint: int
 
         :returns: lines of bytes read from ``rfile``
         :rtype: list[bytes]
@@ -1478,9 +1494,10 @@ class HTTPConnection:
 
         try:
             shutdown(socket.SHUT_RDWR)  # actually send a TCP FIN
+        except errors.acceptable_sock_shutdown_exceptions:
+            pass
         except socket.error as e:
-            # Suppress "client is no longer connected"
-            if e.errno != errno.ENOTCONN:
+            if e.errno not in errors.acceptable_sock_shutdown_error_codes:
                 raise
 
 
@@ -1523,6 +1540,11 @@ class HTTPServer:
     timeout = 10
     """The timeout in seconds for accepted connections (default 10)."""
 
+    expiration_interval = 0.5
+    """The interval, in seconds, at which the server checks for
+    expired connections (default 0.5).
+    """
+
     version = 'Cheroot/{version!s}'.format(version=__version__)
     """A version string for the HTTPServer."""
 
@@ -1592,7 +1614,6 @@ class HTTPServer:
         self.requests = threadpool.ThreadPool(
             self, min=minthreads or 1, max=maxthreads,
         )
-        self.serving = False
 
         if not server_name:
             server_name = self.version
@@ -1769,7 +1790,7 @@ class HTTPServer:
                 info = [(sock_type, socket.SOCK_STREAM, 0, '', bind_addr)]
 
             for res in info:
-                af, socktype, proto, canonname, sa = res
+                af, socktype, proto, _canonname, sa = res
                 try:
                     self.bind(af, socktype, proto)
                     break
@@ -1797,20 +1818,24 @@ class HTTPServer:
 
     def serve(self):
         """Serve requests, after invoking :func:`prepare()`."""
-        self.serving = True
-        while self.ready:
+        while self.ready and not self.interrupt:
             try:
-                self.tick()
+                self._connections.run(self.expiration_interval)
             except (KeyboardInterrupt, SystemExit):
-                self.serving = False
                 raise
             except Exception:
                 self.error_log(
-                    'Error in HTTPServer.tick', level=logging.ERROR,
+                    'Error in HTTPServer.serve', level=logging.ERROR,
                     traceback=True,
                 )
 
-        self.serving = False
+        # raise exceptions reported by any worker threads,
+        # such that the exception is raised from the serve() thread.
+        if self.interrupt:
+            while self._stopping_for_interrupt:
+                time.sleep(0.1)
+            if self.interrupt:
+                raise self.interrupt
 
     def start(self):
         """Run the server forever.
@@ -1829,7 +1854,7 @@ class HTTPServer:
         """Context manager for running this server in a thread."""
         self.prepare()
         thread = threading.Thread(target=self.serve)
-        thread.setDaemon(True)
+        thread.daemon = True
         thread.start()
         try:
             yield thread
@@ -2039,42 +2064,48 @@ class HTTPServer:
 
         return bind_addr
 
-    def tick(self):
-        """Accept a new connection and put it on the Queue."""
-        conn = self._connections.get_conn()
-        if conn:
-            try:
-                self.requests.put(conn)
-            except queue.Full:
-                # Just drop the conn. TODO: write 503 back?
-                conn.close()
-
-        self._connections.expire()
+    def process_conn(self, conn):
+        """Process an incoming HTTPConnection."""
+        try:
+            self.requests.put(conn)
+        except queue.Full:
+            # Just drop the conn. TODO: write 503 back?
+            conn.close()
 
     @property
     def interrupt(self):
         """Flag interrupt of the server."""
         return self._interrupt
 
+    @property
+    def _stopping_for_interrupt(self):
+        """Return whether the server is responding to an interrupt."""
+        return self._interrupt is _STOPPING_FOR_INTERRUPT
+
     @interrupt.setter
     def interrupt(self, interrupt):
-        """Perform the shutdown of this server and save the exception."""
-        self._interrupt = True
+        """Perform the shutdown of this server and save the exception.
+
+        Typically invoked by a worker thread in
+        :py:mod:`~cheroot.workers.threadpool`, the exception is raised
+        from the thread running :py:meth:`serve` once :py:meth:`stop`
+        has completed.
+        """
+        self._interrupt = _STOPPING_FOR_INTERRUPT
         self.stop()
         self._interrupt = interrupt
-        if self._interrupt:
-            raise self.interrupt
 
     def stop(self):  # noqa: C901  # FIXME
         """Gracefully shutdown a server that is serving forever."""
+        if not self.ready:
+            return  # already stopped
+
         self.ready = False
         if self._start_time is not None:
             self._run_time += (time.time() - self._start_time)
         self._start_time = None
 
-        # ensure serve is no longer accessing socket, connections
-        while self.serving:
-            time.sleep(0.1)
+        self._connections.stop()
 
         sock = getattr(self, 'socket', None)
         if sock:
@@ -2100,7 +2131,7 @@ class HTTPServer:
                         host, port, socket.AF_UNSPEC,
                         socket.SOCK_STREAM,
                     ):
-                        af, socktype, proto, canonname, sa = res
+                        af, socktype, proto, _canonname, _sa = res
                         s = None
                         try:
                             s = socket.socket(af, socktype, proto)
